#include<fstream>
#include<iomanip>
#include<cmath>

// Gerate mesh6.poly file for input to DynaMesh for Example 6
// Gerate mesh6.poly file for input to DynaMesh for Example 6
// Gerate mesh6.poly file for input to DynaMesh for Example 6

using namespace std;

#include"svector.C"
#include"smatrix.C"

typedef svector<2> dvector;
typedef smatrix<2> dmatrix;

class edge
{
public:
  edge() {}
  edge(int iia, int iib, dvector xx, bool f, int cc) :
    ia(iia), ib(iib), xmid(xx), flag(f), color(cc) {}

  bool flag;
  int ia,ib, color;
  dvector xmid;
};

const double Pi = 3.1415926535897932385;

const int Level = 3;

// bottom = (0,1,2), lhs = 3, rhs = 4, top = 5

const int N0 = Level*4, N1 = Level*16, N2 = Level*8, N3 = Level*8;
const int N4 = N3, N5 = 2*N2;

const int bc3 = 2, bc4 = 2;                    // triangle messes up
const int bc0 = 1, bc1 = 1, bc2 = 1, bc5 = 1;  // bdy markes with label 0

dvector pts[N0+N1+N2+N3+N4+N5];
edge edges[N0+N1+N2+N3+N4+N5];

void output(int n, dvector pts[], edge edges[]);
dvector rkstepV(dvector x, double h);
dvector perp(dvector v);
dvector rkstepVperp(dvector x, double h);
void velocity6(const double t, const dvector xx, dvector &v, dmatrix &dv);
void velocity6a(const double t, const dvector xx, dvector &v, dmatrix &dv);

int main()
{
  int i, ia, ib, npt = 5, nedge = -1;
  double h, o, r, xmid,ymid, xb,yb, dt, el = 1.0;
  dvector x, xnew;

  cout << "Mesh 6: Level = " << Level 
	<< ", (N0,N1,N2,N3: N4, N5) = (" 
       << N0 << "," << N1 << "," << N2 << "," << N3 << ","
       << N4 << "," << N5 << "," << ")" << endl;

  pts[0] = dvector(0.00, 0.0);
  pts[1] = dvector(0.25, 0.0);
  pts[2] = dvector(0.50, 0.0);
  pts[4] = dvector(0.00, 0.5);

  // first bottom segment

  ia = 0;
  h = (pts[1][0] - pts[0][0]) / N0;

  for(i = 0; i < N0; i++) 
    {
      if(i < N0-1) 
	{
	  ib = ++npt;
	  pts[npt] = dvector((i+1)*h, 0.0);
	}
      else 
        ib = 1;

      edges[++nedge] = edge(ia, ib, dvector(0.0,0.0), false, bc0);

      ia = ib;
    }

  // semi circle

  r = 0.5 * (pts[2][0] - pts[1][0]); 
  o = 0.5 * (pts[2][0] + pts[1][0]); 

  ia = 1;
  h = Pi / N1;

  for(i = 0; i < N1; i++) 
    {
      if(i < N1-1) 
	{
	  ib = ++npt;
	  pts[npt] = dvector(o-r*cos((i+1)*h), r*sin((i+1)*h));
	}
      else
	ib = 2;

      x = dvector(o+r*cos((i+0.5)*h), r*sin((i+0.5)*h));
      edges[++nedge] = edge(ia, ib, x, true, bc1);

      ia = ib;
    }

  // lhs

  ia = 0;
  h = (pts[4][1] - pts[0][1]) / N3;
  x = dvector(0.0,0.0);

  for(i = 0; i < N3; i++) 
    {
      if(i < N3-1) 
	{
	  ib = ++npt;
	  pts[npt] = dvector(0.0, (i+1)*h);
	}
      else
	ib = 4;

      edges[++nedge] = edge(ib, ia, x, false, bc3);
      
      ia = ib;
    }

  // top 

  i = 0;
  ia = 4;

  h = el / N5; 
  dt = h/10;

  xb = h;
  xmid = h / 2;

  x = pts[4];
  
  while(1)
    {
      xnew = rkstepV(x,dt);
      
      if(xnew[0] >= xmid)
	{
	  ymid = x[1] + (xnew[1]-x[1]) * (xmid-x[0])/(xnew[0]-x[0]);
	  xmid += h;
	}
      
      if(xnew[0] >= xb)
	{
	  if(++i == N5) ib = 5;
	  else          ib = ++npt;

	  yb = x[1] + (xnew[1]-x[1]) * (xb-x[0])/(xnew[0]-x[0]);
	  
	  pts[ib] = dvector(xb,yb);
	  
	  edges[++nedge] = edge(ib,ia, dvector(xmid-h,ymid), true, bc5);
	  
	  if(ib == 5) break;

	  ia = ib;
	  xb += h;
	}

      x = xnew;
    }

  // rhs 

  i = 0;
  ia = 5;

  h = pts[5][1] / N4; 
  dt = h / 10;

  yb   = pts[5][1] - h;
  ymid = pts[5][1] - h/2;

  x = pts[5];
  
  while(1)
    {
      xnew = rkstepVperp(x,dt);
      
      if(xnew[1] <= ymid)
	{
	  xmid = x[0] + (xnew[0]-x[0]) * (ymid-x[1])/(xnew[1]-x[1]);
	  ymid -= h;
	}
      
      if(xnew[1] <= yb)
	{
	  if(++i == N4) ib = 3;
	  else          ib = ++npt;

	  xb = x[0] + (xnew[0]-x[0]) * (yb-x[1])/(xnew[1]-x[1]);
	  
	  pts[ib] = dvector(xb,yb);
	  
	  edges[++nedge] = edge(ib,ia, dvector(xmid,ymid+h), true, bc4);
	  
	  if(ib == 3) break;

	  ia = ib;
	  yb -= h;
	}

      x = xnew;
    }
      
  // second bottom segment

  ia = 2;

  o = pts[2][0];
  h = (pts[3][0] - pts[2][0]) / N2;

  x = dvector(0.0,0.0);

  for(i = 0; i < N2; i++) 
    {
      if(i < N2-1) 
	{
	  ib = ++npt;
	  pts[npt] = dvector(o+(i+1)*h, 0.0);
	}
      else
	ib = 3;

      edges[++nedge] = edge(ia,ib, x, false, bc2);

      ia = ib;
    }

  output(N0+N1+N2+N3+N4+N5, pts, edges);

  return(0);
}

void output(int n, dvector pts[], edge edges[])
{
  ofstream fout("mesh6.poly");
  
  fout.setf(ios::fixed);

  if(!fout.is_open())
    {
      cerr << "Can't open output file -- terminating" << endl;
      throw("Can't open output file");
    }
  
  fout << n << "  " << 2 << "  " << 0 << "  " << 0 << endl;
  
  for(int i = 0; i < n; i++)
    {
      fout << setw(2) << i << "  " 
	   << setw(10) << pts[i][0] << "  "
	   << setw(10) << pts[i][1] << endl;
    }

  fout << n << "  " << 1 << endl;

  for(int i = 0; i < n; i++)
    {
      fout << setw(2) << i << "  " 
	   << setw(2) << edges[i].ia << "  " 
	   << setw(2) << edges[i].ib << "  " 
	   << setw(2) << edges[i].color;

     
      //      if(edges[i].flag) 
      //	fout << " " << setw(10) << edges[i].xmid[0] 
      //	     << " " << setw(10) << edges[i].xmid[1];
     
      fout << endl;
    }

  fout << 0 << endl;

  fout << 1 << " " << 1 << endl;       // No. colored pts, No. colors
  fout << "1 0.25 0.25 1" << endl;     // Point No. (x,y) color

  fout.close();

  ofstream fqual("mesh6.qual");
  
  fqual.setf(ios::fixed);

  if(!fqual.is_open())
    {
      cerr << "Can't open quality output file -- terminating" << endl;
      throw("Can't open output file");
    }

  fqual << 1 << endl;    // one color
  fqual << 1 << " " << 30.0 << " " << 140.0 << " "   // color min/max angle
	<< pts[1][0]*pts[1][0] / (N0*N0) << " "      // max area
	<< 0.0 << endl;                              

  fqual.close();
}

dvector rkstepV(dvector x, double h)
{
  double ho2 = h / 2.0, t=0.0;
  dvector v, k1,k2,k3,k4;
  dmatrix dv;

  velocity6(t, x, v, dv); k1 = h * v;
  
  velocity6(t, x + 0.5*k1, v, dv); k2 = h * v;
  velocity6(t, x + 0.5*k2, v, dv); k3 = h * v;
  
  velocity6(t, x + k3, v, dv); k4 = h * v;
  
  return(x + (1/6.0)*(k1 + 2*(k2 + k3) + k4));
}

dvector perp(dvector v) {return( dvector(v[1], -v[0]) );}

dvector rkstepVperp(dvector x, double h)
{
  double ho2 = h / 2.0, t=0.0;
  dvector v, k1,k2,k3,k4;
  dmatrix dv;

  velocity6(t, x, v, dv); k1 = h * perp(v);

  velocity6(t, x + 0.5*k1, v, dv); k2 = h * perp(v);
  velocity6(t, x + 0.5*k2, v, dv); k3 = h * perp(v);
     
  velocity6(t, x + k3, v, dv); k4 = h * perp(v);

  return(x + (1/6.0)*(k1 + 2*(k2 + k3) + k4));
}

void velocity6a(const double t, const dvector xx, dvector &v, dmatrix &dv)
{
  v = dvector(1,0);
}

void velocity6(const double t, const dvector xy, dvector &v, dmatrix &dv)
{
  double re = 1, a = 0.125, U = 5.0;
  double gamma = 0.57721566490153286061, C = 2*U/(0.5-gamma-log(re/4));


  double x = xy[0] - 0.375, y = xy[1], rsq = x*x + y*y; 
  
  if(rsq <= a*a) 
    {
      v.zero();
      dv.zero();

      return;
    }

  double r = sqrt(rsq), aa = a/r, xx = x/r, yy = y/r;
  double aasq = aa*aa, xxsq = xx*xx, yysq = yy*yy, m[2][2];

  v = dvector(U + 
	       0.5*C*(gamma - 0.5 + log(re/(4*aa)) + 0.5*(1-aasq)*(1-2*xxsq)),
	      -0.5*C*(1-aasq)*xx*yy );

  m[0][0] = xx*( 1-3*aasq+4*xxsq*aasq-2*xxsq);
  m[0][1] = yy*(-1-  aasq+4*xxsq*aasq-2*xxsq);
  m[1][0] = yy*( 1-  aasq+4*xxsq*aasq-2*xxsq);
  m[1][1] = xx*( 1-  aasq+4*yysq*aasq-2*yysq);

  dv = (-0.5*C/r) * dmatrix(m);

  return;
}
