#include"triBst.C"

class felement : public gelement {
public:
  felement(const gelement & ge) : gelement(ge) {}

  int ndof() { return(Nphi);}

  int dof[Nphi];
  svector<Ndim> pts[Nphi];
private:
};

class bfelement : public bgelement {
public:
  bfelement(const bgelement & bge) : bgelement(bge) {}
  
  int ndof() { return(Nphib); }
  
  int dof[Nphib];
  svector<Ndim> pts[Nphib];
private:  
};

class feMesh : public gmMesh
{
public:
  feMesh(string);
  
  ~feMesh() { bstfree(bst); }
  
  int band();
  int ndof() {return(nd);}
  
  class feiterator : public gmMesh::geiterator
  {
  public:
    feiterator(const geiterator & it, feMesh * f);
    felement operator*();
    
  private:
    feMesh * fm;
  };
  
  feiterator febegin();
  feiterator feend();
  
  class bfeiterator : public gmMesh::bgeiterator
  {
  public:
    bfeiterator(const bgeiterator & it, feMesh * f);
    bfelement operator*();
    
  private:
    feMesh * fm;
  };
  
  bfeiterator bfebegin();
  bfeiterator bfeend();
  
private:
  mutable bnode *bst;
  
  int nd;
  
  int key2dof(int * key);
  
  felement geToFe(const gelement &);

  bfelement bgeToBfe(const bgelement &);

  friend ostream& operator<<(ostream& out, feMesh &dd);
};

feMesh::feMesh(string ss)
  :  gmMesh(ss), bst(NULL)
{
  int j, key[2];

  cout << "Initializing feMesh()" << endl;

 for (gmMesh::geiterator geit = gebegin(); 
      geit != geend(); geit++ )
 {
     gelement ee = *geit;

    for(j = 0; j < ee.master->nn; j++)
      {
      key[0] = key[1] = ee.nodes[j];

      bstadd(&bst, bstsort(key));
      }

    #ifdef QuadraticElements
    for(j = 0; j < ee.master->ne; j++)
      {
      key[0] = ee.nodes[ ee.master->edges[j][0] ];
      key[1] = ee.nodes[ ee.master->edges[j][1] ];

      bstadd(&bst, bstsort(key));
      }
    #endif
    }

  nd = bstset(bst);        // Assign dof numbers

  cout << "Number of degrees of freedom = ................... " << nd
       << endl;

  cout << "Leaving constructor" << endl;

}

int feMesh::band()
{
  int bb = 0, dmin, dmax;

  for (feMesh::feiterator feit = febegin(); 
      feit != feend(); feit++ )
    {
    felement ee = *feit;

    dmin = dmax = ee.dof[0];
   
    for(int j = 0; j < ee.ndof(); j++)
      {
      if(dmin > ee.dof[j]) dmin = ee.dof[j];
      if(dmax < ee.dof[j]) dmax = ee.dof[j];
      }
    
    if(bb < (dmax-dmin+1)) bb = dmax-dmin+1;
    }
 
  return(bb);
}

int feMesh::key2dof(int * key) {
    return bstdat(&bst, bstsort(key));
}

felement feMesh::geToFe(const gelement &ge)
{
  felement fe(ge);
    
  // set dof

  int j, key[2], nn = fe.master->nn;
  
  for(j = 0; j < nn; j++)
    {
      key[0] = key[1] = fe.nodes[j];
      
      fe.dof[j] = key2dof(key);
    }
  
  #ifdef QuadraticElements
  for(j = 0; j < fe.master->ne; j++)
    {
      key[0] = fe.nodes[ fe.master->edges[j][0] ];
      key[1] = fe.nodes[ fe.master->edges[j][1] ];
      
      fe.dof[j+nn] = key2dof(key);
    }
  #endif
  
  for(j = 0; j < nn; j++)
    {
      fe.pts[j] = fe.xx[j];
    }
  
  #ifdef QuadraticElements
    for(j = 0; j < fe.master->ne; j++)
      {
	fe.pts[j+nn] = 0.5 * (fe.xx[ fe.master->edges[j][0] ] +
			      fe.xx[ fe.master->edges[j][1] ]);
      }
  #endif

    return(fe);
}

bfelement feMesh::bgeToBfe(const bgelement &ge)
{
  bfelement fe(ge);
    
  // set fe.dof
  
  int j, key[2], nn = fe.master->nn;
  
  for(j = 0; j < nn; j++)
    {
      key[0] = key[1] = fe.nodes[j];
      
      fe.dof[j] = key2dof(key);
    }
  
  #ifdef QuadraticElements
  for(j = 0; j < fe.master->ne; j++)
    {
      key[0] = fe.nodes[ fe.master->edges[j][0] ];
      key[1] = fe.nodes[ fe.master->edges[j][1] ];
      
      fe.dof[j+nn] = key2dof(key);
    }
  #endif
  
  //set fe.pts
  
  for(j = 0; j < nn; j++)
    {
      fe.pts[j] = fe.xx[j];
    }
  
  #ifdef QuadraticElements
  for(j = 0; j < fe.master->ne; j++)
    {
      fe.pts[j+nn] = 0.5 * (fe.xx[ fe.master->edges[j][0] ] +
			    fe.xx[ fe.master->edges[j][1] ]);
    }
  #endif
    
  return fe;
}

ostream& operator<<(ostream& out, feMesh &dd)
{
  out << "Finite Element Mesh:" << endl;

  return(out);
}

// Fe iterator

feMesh::feiterator::feiterator(const feMesh::geiterator & it, feMesh * f) 
  : gmMesh::geiterator(it)
{ 
    fm = f;
}
    
felement feMesh::feiterator::operator*() 
{
  return fm->geToFe(this->geiterator::operator*());
}

feMesh::feiterator feMesh::febegin() 
{
    return feiterator(gebegin(),this);
}

feMesh::feiterator feMesh::feend() 
{
    return feiterator(geend(),this);
}

// be iterator

feMesh::bfeiterator::bfeiterator(const feMesh::bgeiterator & it, feMesh * f) 
  : gmMesh::bgeiterator(it)
{ 
    fm = f;
}

bfelement feMesh::bfeiterator::operator*() 
{
  return fm->bgeToBfe(this->bgeiterator::operator*());
}

feMesh::bfeiterator feMesh::bfebegin() {
    return bfeiterator(bgebegin(),this);
}

feMesh::bfeiterator feMesh::bfeend() {
    return bfeiterator(bgeend(),this);
}
