// Set up mesh with elements affine equivalent to the unit tetrahedra

#include<string>
#include<fstream>
#include<map>
#include<vector>
#include"foreach.hpp"

#include "allh.h"

#include "gelement3d.C"

class dir3Mesh
{
public:
    dir3Mesh(string); 
    dir3Mesh(const dir3Mesh &mm) 
	{
	    cerr << "dir3Mesh(): Don't copy me, deep copy nyi" << endl;
	    throw("don't copy a dir3Mesh");
	}
    dir3Mesh() 
	{
	    cerr << "dir3Mesh(): Default constructor called, nyi" << endl;
	    throw("bad dir3Mesh constructor");
	}
    
    ~dir3Mesh() { delete C; return; }
    
    void print(ostream& dump);  // FIXME: implement this
    void print(string fileName);  // FIXME: implement this
    
    //  void locate(const dvector&, gelement&, dvector&) const;
    
    class geiterator {
    public:
	geiterator();
	geiterator(const Mesh::siterator & s, PLC * C);
	void operator++();
	void operator++(int);
	geiterator& operator=(const geiterator & it);
	bool operator==(const geiterator & it);
	bool operator!=(const geiterator & it);
        gelement operator*();
	
    private:
	PLC * C;
	Mesh::siterator sit;
    };
    
    geiterator gebegin();
    geiterator geend();
    
    class bgeiterator {
    public:
	bgeiterator();
	bgeiterator(const Mesh::siterator & s, PLC * C, int face);
	void operator++();
	void operator++(int);
	bgeiterator& operator=(const bgeiterator & it);
	bool operator==(const bgeiterator & it);
	bool operator!=(const bgeiterator & it);
        bgelement operator*();
	
    private:
	PLC * C;
	int face;
	Mesh::siterator sit;
    };

    bgeiterator bgebegin();
    bgeiterator bgeend();
    
private:
    PLC * C;
    
    Mesh::Iterator sit; // tet iterator
    
    int face;
    Mesh::Iterator sitb; // tri iterator
};

dir3Mesh::dir3Mesh(string meshFile) 
{
  cout << "Initializing dir3Mesh(" << meshFile << "):" 
       << " (reads .plc file)" << endl;
  
  C = new PLC();
  // FIXME
  // set up configuration here
  // to supress unneeded output
  C->ReadAndRefine(meshFile);

  return;
}

dir3Mesh::geiterator::geiterator() {
    C = NULL;
}

dir3Mesh::geiterator::geiterator(const Mesh::siterator & s, PLC * complex) {
    sit = s;
    C = complex;
}

void dir3Mesh::geiterator::operator++() {
    (*this)++;
}

void dir3Mesh::geiterator::operator++(int) {
    sit++;
}

dir3Mesh::geiterator& dir3Mesh::geiterator::operator=(const geiterator & it) {
    C = it.C;
    sit = it.sit;
    return *this;
}

bool dir3Mesh::geiterator::operator==(const geiterator & it) {
    if (C != it.C) {
	return false;
    }
    if (C == NULL) {
	return true;
    }
    if (sit == it.sit) {
	return true;
    }
    return false;
}

bool dir3Mesh::geiterator::operator!=(const geiterator & it) {
    return !(*this == it);
}

gelement dir3Mesh::geiterator::operator*() {

    gelement ee;
    Simplex * s = *sit;

    ee.master = &LinearTetrahedron;

    for(int j = 0; j < 4; j++) 
    {
	ee.xx[j] = svector<3>(C->P->V[s->v[j]].xyz);	
	ee.nodes[j] = s->v[j];
    }
    return ee;
}


dir3Mesh::geiterator dir3Mesh::gebegin() {
    return geiterator(C->M->sbegin(), C);
}

dir3Mesh::geiterator dir3Mesh::geend() {
    return geiterator(C->M->send(), C);
}

dir3Mesh::bgeiterator::bgeiterator() {
    C = NULL;
    face = -1;
}

dir3Mesh::bgeiterator::bgeiterator(const Mesh::siterator & s, PLC * complex, int f) {
    sit = s;
    C = complex;
    face = f;
}

void dir3Mesh::bgeiterator::operator++() {
    (*this)++;
}

void dir3Mesh::bgeiterator::operator++(int) {

    int found = 0;
    int reset = 0;
    while (found == 0) 
    {
	while (face < (int)C->M->children.size() && 
	       (face < 0 || C->M->children[face]->dim != 2) ) 
	{
	    face++;
	    reset = 1;
	    
	}
	if (face >= (int)C->M->children.size()) 
	{
	    return;
	}
	Mesh * m = C->M->children[face];
	
	if (reset == 1) {
	    sit = m->sbegin();
	} else {
	    sit++;
	}

	found = 1;
	while(sit != m->send() && 
	      m->Exterior(*sit) )
	{
	    sit++;
	}
	if (sit==m->send()) 
	{
	    // done with this face... move to the next one
	    found = 0;
	    reset=1;
	    face++;
	}
	if (face >= (int)C->M->children.size()) 
	{
	    return;
	}
    }
}

dir3Mesh::bgeiterator& dir3Mesh::bgeiterator::operator=(const bgeiterator & it) {
    C = it.C;
    face = it.face;
    sit = it.sit;
    return *this;
}

bool dir3Mesh::bgeiterator::operator==(const bgeiterator & it) {
    if (C != it.C) {
	return false;
    }
    if (C == NULL) {
	return true;
    }
    
    // .end() iterator
    if (face == it.face && face >= (int)C->M->children.size()) {
	return true;
    }

    if (face == it.face) {
	if (sit == it.sit) {
	    return true;
	}
    }
    return false;
}

bool dir3Mesh::bgeiterator::operator!=(const bgeiterator & it) {
    return !(*this == it);
}

bgelement dir3Mesh::bgeiterator::operator*() {

    bgelement be;
    Simplex * s = *sit;

    be.master = &LinearTriangle;

    for(int j = 0; j < 3; j++) 
    {
	be.xx[j] = svector<3>(C->P->V[s->v[j]].xyz);	
	be.nodes[j] = s->v[j];
    }

    be.bctype = C->M->children[face]->id;

    return be;
}

dir3Mesh::bgeiterator dir3Mesh::bgebegin() {
    Mesh::siterator it;
    bgeiterator bgeit(it,C,-1);
    bgeit++;
    return bgeit;;
}

dir3Mesh::bgeiterator dir3Mesh::bgeend() {
    return bgeiterator(C->M->send(), C, C->M->children.size());
}
