// Set up triangular element with quadratic basis functions

#include<string>
#include<fstream>
#include<vector>

#include "gelement2d.C"

class triangleMesh
{
public:
  triangleMesh(string);                        // read mesh.poly file
  triangleMesh(const triangleMesh &mm) 
  {
    cerr << "triangleMesh(): don't copy a triangleMesh" << endl;
    throw("copying a triangleMesh");
  }
    
    triangleMesh() 
    {
	cerr << "triangleMesh(): Default constructor called, nyi" << endl;
	throw("bad triangleMesh constructor");
    }

  ~triangleMesh() {return;}

  void print(ostream& dump);

  int nelem;
  int nodes;
  int nbelem;

  bool checkBdyOrientation(svector<2> cc);
    
  //  void locate(const svector<2>&, gelement&, svector<2>&) const;

  //gelement element(int i);
  //belement boundary(int i);

    class geiterator {
    public:
	geiterator();
	geiterator(int l, triangleMesh * g);
	void operator++();
	void operator++(int);
	geiterator& operator=(const geiterator & it);
	bool operator==(const geiterator & it);
	bool operator!=(const geiterator & it);
        gelement operator*();

    private:
	int loc;
	triangleMesh * gm;
    };

    geiterator gebegin();
    geiterator geend();

    class bgeiterator {
    public:
	bgeiterator();
	bgeiterator(int l, triangleMesh * g);
	void operator++();
	void operator++(int);
	bgeiterator& operator=(const bgeiterator & it);
	bool operator==(const bgeiterator & it);
	bool operator!=(const bgeiterator & it);
        bgelement operator*();

    private:
	int loc;
	triangleMesh * gm;
    };

    bgeiterator bgebegin();
    bgeiterator bgeend();



private:
  vector<svector<2> > vertices;       // vertex coordinates

  vector<intArray<2> > segments;      // boundary edges
  vector<int> segmentColors;

  vector<intArray<3> > triangles;     // triangles

  void input(string meshFile);         // auxillary function to read data
};

triangleMesh::triangleMesh(string meshFile) 
{
  cout << "Initializing triangleMesh(" << meshFile << "):" 
       << " (reads .node, .poly, and .ele files)" << endl;

  input(meshFile);

  cout << "Number of vertices read in = ..................... " << nodes
       << endl;
  cout << "Number of boundary segments = .................... " << nbelem 
       << endl;
  cout << "Number of triangles = ............................ " << nelem 
       << endl;

  return;
}


/*
belement triangleMesh::boundary(int i)
{
  int j, *ss = segments[i];
  belement be;

  assert((-1 < i) && (i < nbelem));

  be.bctype = segmentColors[i];

  be.master = &LinearSegment;

  for(j = 0; j < 2; j++) 
    {    
    be.nodes[j] = ss[j];

    be.xx[j] = vertices[ ss[j] ];
    }

  return(be);
}
*/

triangleMesh::geiterator::geiterator() {
    loc = 0;
    gm = NULL;
}

triangleMesh::geiterator::geiterator(int l, triangleMesh * g) {
    loc = l;
    gm = g;
}


void triangleMesh::geiterator::operator++() {
    (*this)++;
}

void triangleMesh::geiterator::operator++(int) {
    loc++;
}

triangleMesh::geiterator& triangleMesh::geiterator::operator=(const geiterator & it) {
    loc = it.loc;
    gm = it.gm;
    return *this;
}

bool triangleMesh::geiterator::operator==(const geiterator & it) {
    if (gm != it.gm || loc != it.loc) {
	return false;
    }
    return true;
}

bool triangleMesh::geiterator::operator!=(const geiterator & it) {
    return !(*this == it);
}

gelement triangleMesh::geiterator::operator*() {

    gelement ee;
    int j, *tt = gm->triangles[loc]; 

    ee.master = &LinearTriangle;
    
    for(j = 0; j < 3; j++) 
    {
	ee.xx[j] = gm->vertices[tt[j]];
	ee.nodes[j] = tt[j];
    }
    
    return ee;
}


triangleMesh::geiterator triangleMesh::gebegin() {
    return geiterator(0, this);
}

triangleMesh::geiterator triangleMesh::geend() {
    return geiterator(nelem, this);
}

triangleMesh::bgeiterator::bgeiterator() {
    loc = 0;
    gm = NULL;
}

triangleMesh::bgeiterator::bgeiterator(int l, triangleMesh * g) {
    loc = l;
    gm = g;
}

void triangleMesh::bgeiterator::operator++() {
    (*this)++;
}

void triangleMesh::bgeiterator::operator++(int) {
    loc++;
}

triangleMesh::bgeiterator& triangleMesh::bgeiterator::operator=(const bgeiterator & it) {
    loc = it.loc;
    gm = it.gm;
    return *this;
}

bool triangleMesh::bgeiterator::operator==(const bgeiterator & it) {
    if (loc != it.loc || gm != it.gm) {
	return false;
    }
    return true;
}

bool triangleMesh::bgeiterator::operator!=(const bgeiterator & it) {
    return !(*this == it);
}

bgelement triangleMesh::bgeiterator::operator*() {

  int j, *tt = gm->segments[loc];
  bgelement be;

  be.bctype = gm->segmentColors[loc];

  be.master = &LinearSegment;

  for(j = 0; j < 2; j++) 
  {
      be.xx[j] = gm->vertices[tt[j]];
      
      be.nodes[j] = tt[j];
  }  
  return be;

}


triangleMesh::bgeiterator triangleMesh::bgebegin() {
    return bgeiterator(0,this);
}


triangleMesh::bgeiterator triangleMesh::bgeend() {
    return bgeiterator(nbelem,this);
}



void triangleMesh::print(ostream& dump = cout)
{
  dump.setf(ios::fixed,ios::floatfield);

  dump << "Number of elements = " << nelem << endl;

  //for(int i = 0; i < nelem; i++)
  //  dump << "Element[" << i << "]" << endl << element(i) << endl;

  dump << endl << "Number of boundary elements = " << nbelem << endl;

  //for(int i = 0; i < nbelem; i++)
  //  dump << "Boundary[" << i << "]" << boundary(i) << endl;

  return;
}

void triangleMesh::input(string meshFile)
{
  int nAttr, nBmkr, color, dim, index, i0,i1, i2;
  double xx,yy;

  ifstream in((meshFile + ".node").c_str());

  if( !in.is_open() ) 
    {
      cerr << "can't open file:" << meshFile + ".node" << endl;

      throw("Can't open file");
    }

  // Read in the vertices

  in >> nodes >> dim >> nAttr >> nBmkr;

  assert(dim == 2);

  assert(nodes > 0);

  vertices.resize(nodes);

  for(int i = 0; i < nodes; i++)
    {
      in.ignore(256,'\n');

      in >> index;

      if(index != i)
	cerr << "strToMesh(): Warning: data in .poly may be bad" << endl;

      in >> xx >> yy;

      vertices[i] = svector<2>(xx,yy);

      // ignore the rest
    }

  in.close();

  // Read in segments from the .poly file

  in.open((meshFile + ".poly").c_str());

  if( !in.is_open() ) 
    {
      cerr << "can't open file:" << meshFile + ".poly" << endl;

      throw("Can't open file");
    }

  in >> i0 >> dim >> nAttr >> nBmkr;

  assert(dim == 2);

  assert(i0 == 0);    // i0 = 0 means vertices in .node file
    
  in.ignore(256,'\n');
  
  in >> nbelem >> nBmkr;

  color = 0;      // default color if nBmkr = 0 (Dirichlet bc)

  segments.resize(nbelem);
  segmentColors.resize(nbelem);

  for(int i = 0; i < nbelem; i++)
    {
      in.ignore(256,'\n');

      in >> index;

      if(index != i) cerr << "Warning: data in file may be bad" << endl;

      in >> i0 >> i1;

      assert((-1 < i0) && (i0 < nodes));
      assert((-1 < i1) && (i1 < nodes));

      segments[i][0] = i0;
      segments[i][1] = i1;

      if(nBmkr > 0) in >> color;

      assert((color==1) || (color==2));

      segmentColors[i] = color-1;   // triangle messes up zero labels
    }

  in.close();

  // Read in the triangles

  in.open( (meshFile + ".ele").c_str() );

  if( !in.is_open() ) 
    {
      cerr << "can't open file:" << meshFile + ".ele" << endl;

      throw("Can't open file");
    }

  in >> nelem >> i0 >> nAttr;

  triangles.resize(nelem);

  for(int i = 0; i < nelem; i++)
    {
      in.ignore(256,'\n');

      in >> index;

      if(index != i) cerr << "Warning: data in file may be bad" << endl;

      in >> i0 >> i1 >> i2;

      assert((-1 < i0) && (i0 < nodes));
      assert((-1 < i1) && (i1 < nodes));
      assert((-1 < i2) && (i2 < nodes));

      triangles[i][0] = i0;
      triangles[i][1] = i1;
      triangles[i][2] = i2;

      // ignore the rest
    }

  in.close();

  return;
}

bool triangleMesh::checkBdyOrientation(svector<2> cc)
{
  // loop over boundary elements and check that they
  // go counterclockwise around the point cc

  // for this to work the domain must be star shaped wrt cc

  bool test = true;
  double cross;
  svector<2> v0, v1;

  for(int i = 0; i < nbelem; i++)
    {
      v0 = vertices[ segments[i][0]  ] - cc;
      v1 = vertices[ segments[i][1] ] - cc;

      cross = v0[0]*v1[1] - v0[1]*v1[0];

      if(cross < 0)   // fix thes
	{
	  test = false;

	  int temp = segments[i][0];

	  segments[i][0] = segments[i][1];
	  segments[i][1] = temp;
	}   
    }

  return(test);
}

