#include"poisson.h"

// Sample program to solve Poisson's equation
// Sample program to solve Poisson's equation
// Sample program to solve Poisson's equation

double (*exact)(dvector xx, dvector &du);
void   (*coeff)(dvector xx, dmatrix &aa, double &b);
double (*rhs)(dvector xx);

int main(int argc, char *argv[])
{
  int eg;
  clock_t t00 = clock(), t0 = t00;
  double ae[Nphi][Nphi],   fe[Nphi];
  double ab[Nphib][Nphib], fb[Nphib];

 feMV mvclass(argc, &argv);   // the constructor and destructor will be called

 if(argc < 2)
   {
     cerr << "Usage: " << argv[0] << " file_name; terminating" << endl;

     throw("no argument");
   }

 if(argc > 2)
   {
   istringstream ist(argv[2]);

   if( !(ist >> eg) )
     {
     cerr << "Invalid value for eg:" << argv[2] << ", terminating" << endl;
     throw("Invalid input");
     }

   if((eg < 0) || (eg > EgMax)) 
     {
     cerr << "Invalid value for eg:" << eg << ", terminating" << endl;
     throw("Invalid input argument");
     }
   }
 else
   {
   eg = 0;
   }

 // Evaluate FE basis functions at the Gauss points

 initializeBasis();

 // Set the problem data

 (Examples[eg].initialize)();
 exact   = Examples[eg].exact;
 coeff   = Examples[eg].coeff;
 rhs     = Examples[eg].rhs;

 cout << "Sample finite element code:" << endl;
 cout << "Solution of " << Ndim << "d Poisson equation" << endl;
 cout << "Example " << eg << ": " << Examples[eg].desc << endl;

 // Construct the mesh and dof

 feMesh mm(argv[1]);

 // triangle may not return a consistent boundary orientations

#if NDIM == 2
 if( !mm.checkBdyOrientation(dvector(0.375,0.25)) )
   cerr << "main(): boundary orientation test failed; fixup attempted\n";
#endif

 // Allocate the global matrix and vectors

 FeMdata mdata;
 mdata.nvar = mm.ndof();
 mdata.band = mm.band();
 mdata.nzero = Nphi * 5 *(Ndim-1);    // Estimate of non-zero entries/row

 feMV::matrix A(mdata); 
 feMV::vector U(mm.ndof());
 feMV::vector F(mm.ndof());

 cout << "Number of basis functions per element = .......... " << Nphi << endl;
 cout << "Number of unknowns = ............................. " 
      << mdata.nvar << endl;
 cout << "Half band width = ................................ " 
      << mdata.band << endl;
 cout << "Estimated size of big matrix = " 
      << 8.0 * mdata.nvar * mdata.nzero / 1.0e06 << " Mbytes" << endl;

 cout << "Time to initialize mesh and matrix classes ....... "
      << (clock() - t0) / (double) CLOCKS_PER_SEC << endl;
 t0 = clock();

 // Assemble the finite elements

 A.zero();    // Zero the Global  Matricies. 
 F.zero();

 int nelem = 0;
 
 for (feMesh::feiterator feit = mm.febegin(); 
      feit != mm.feend(); feit++ )
   {
     nelem ++;
     felement ee = *feit;
     elment(ee.xx, ae, fe);
     A.assemble(Nphi, ee.dof, (double *)ae, F, fe);
   }
 
 cout << "Number of elements = ............................. " 
      <<  nelem << endl;
 cout << "Time to assemble element matricies ............... "
      << (clock() - t0) / (double) CLOCKS_PER_SEC << endl;
 t0 = clock();
 
 // Specify boundary conditions

 nelem = 0;	
  
 for (feMesh::bfeiterator bfeit = mm.bfebegin(); 
      bfeit != mm.bfeend(); bfeit++) 
   {
     nelem ++;
     bfelement be = *bfeit;     
     bndyc(be, ab, fb);   
     A.assemble(Nphib,be.dof,(double *)ab, F, fb);  
   }

 cout << "Number of boundary elements = .................... " 
      <<  nelem << endl;
 cout << "Time to assemble boundary element matrices ....... "
      << (clock() - t0) / (double) CLOCKS_PER_SEC << endl;
 t0 = clock();

 // solve the linear system and compute the norm

 A.solve(U,F);   // Solve the system of equations 

 cout << "Time to solve linear equations ................... "
      << (clock() - t0) / (double) CLOCKS_PER_SEC << endl;
 t0 = clock();

 norm(U, mm);

 cout << "Time to compute norms ............................ "
      << (clock() - t0) / (double) CLOCKS_PER_SEC << endl;
 t0 = clock();

 output(U);    // Print out the solution.       

 cout << "Total CPU time ................................... "
      << (clock() - t00) / (double) CLOCKS_PER_SEC << endl;

 return(0);
}

void elment(dvector xy[Nphi], double a[Nphi][Nphi] , double f[Nphi])
{
 int i, j, ig;
 double det, weight, bcoeff, frhs;
 dvector xx, dphidx[Nphi];
 dmatrix dxds, dsdx, acoeff;

// zero the element matrix and vector

 for(i = 0; i < Nphi; i++)
   {
   f[i] = 0.0;

   for(j = 0; j < Nphi; j++) a[i][j]  = 0.0;
   }

// quadratre loop

 for(ig = 0; ig < Ng; ig++)
   {
   // Calculate the Transormation Matrix

   xx.zero();      //  Zero out xx = (x,y)
   dxds.zero();

   for(i = 0; i < Nchi; i++) 
     {
       xx += Chi[ig][i] * xy[i];

       dxds += dmatrix(xy[i], Dchi[ig][i]);   // outer product
     }

   dsdx = dxds.inverse(det);    // Invert dxds and return the Jacobian

   if(det == 0.0) goto error;
  
   weight = Gw[ig] * fabs(det);

   dsdx = dsdx.transpose();
 
   for(i = 0; i < Nphi; i++) dphidx[i] = dsdx * Dphi[ig][i];

   frhs = rhs(xx);
   coeff(xx, acoeff, bcoeff);

   for(i = 0; i < Nphi; i++)
     {
     f[i]  += frhs * Phi[ig][i] * weight;

     for(j = 0; j < Nphi; j++)
       a[i][j] += ( dphidx[i].dot(acoeff * dphidx[j]) 
		  +  bcoeff * Phi[ig][i]*Phi[ig][j] ) * weight;
     }
   }

 return;

 error:

 cerr << "Error in elment -- zero Jacobian" << endl;
 throw("Negative Jacobian");
}

void bndyc(bfelement be, double ab[Nphib][Nphib], double fb[Nphib])
{
  int i;
  double big = 1.0e+50;

// zero the matrix and vector

  assert(Nchib == be.master->nn);

 for(i = 0; i < Nphib; i++)
   {
   fb[i] = 0.0;

   for(int j = 0; j < Nphib; j++) ab[i][j]  = 0.0;
   }

 switch(be.bctype%2)   // be.bctype = FaceId (even Dirichlet, odd Neumann)
   {
   case 0:             // type = 0  specifies boundary data on the side

     for(i = 0; i < Nphib; i++)
       {
       ab[i][i] += big;

       fb[i] += big * dirichlet(be.pts[i]);
       }
     break;

     case 1:         // type = 1 specifies Neumann  data on a side 
       {
       double det, weight;
       dvector xx, normal;

       bmatrix dxds;       // 2x1 or 3x2 matrix for 2d/3d
       
       for(int ig = 0; ig < Ngb; ig++)
	 {
	 xx.zero();
	 dxds.zero();
	 
	 for(i = 0; i < Nchib; i++)
	   {
	   xx   +=  Chib[ig][i] * be.xx[i];
	   dxds += bmatrix(be.xx[i], Dchib[ig][i]);  // outer product
	   }
	 
	 normal = dxds.normal(det);   // det = area element         
	 
	 weight = det * Gwb[ig];

	 double g = neumann(xx,normal);

	 for(i = 0; i < Nphib; i++) fb[i] += Phib[ig][i] * g * weight;
	 }
       
       break;
       }

   default:
     cout << "Warning -- Boundary type " << be.bctype  << " ignored" << endl;
   }

 return;
}

double dirichlet(dvector xx)
{
 dvector dudx;
 return( exact(xx,dudx) );
}

double neumann(dvector xx, dvector normal)
{
 dvector dudx;

 exact(xx,dudx);
 return( dudx.dot(normal) ); 
}

void norm(feMV::vector& uu, feMesh &mm)
{
  int i, ig;
 dvector dphidx[Nphi], xx, du, duh;
 dmatrix dsdx, dxds;

 double det, u, uh, ue[Nphi], weight;
 double norm[3], dnorm[3], hnorm[3];

 for(i = 0; i < 3; i++)
   {
   norm[i] = 0.0;
   dnorm[i] = 0.0;
   }

 for (feMesh::feiterator feit = mm.febegin(); 
      feit != mm.feend(); feit++ )
   {
   felement ee = *feit;

   for(ig = 0; ig < Ng ; ig++)
     {
     // Calculate the Transormation Matrix

     xx.zero();
     dxds.zero();
	
     for(i = 0; i < Nchi; i++)
       {
	 xx += Chi[ig][i] * ee.xx[i];
	    
	 dxds += dmatrix(ee.xx[i], Dchi[ig][i]);  // outer product
       }
	
     dsdx = dxds.inverse(det);
	
     if(det == 0.0) goto error;
	
     weight = fabs(det) * Gw[ig];
	
     dsdx = dsdx.transpose();
	
     for(i = 0; i < Nphi; i++)  dphidx[i] = dsdx * Dphi[ig][i];

     u = exact(xx, du);                  /* Get the exact solution.  */

     uh = 0.0;
     duh.zero();

     uu.extract(ee.ndof(), ee.dof, ue);

     for(i = 0; i < Nphi; i++)
       {
       uh  += ue[i] * Phi[ig][i];
       duh += ue[i] * dphidx[i];
       }

     norm[0] += u * u * weight;
     norm[1] += uh*uh * weight;
     norm[2] += (u-uh)*(u-uh) * weight;

     dnorm[0] +=  du.dot(du)  * weight;
     dnorm[1] += duh.dot(duh) * weight;
     dnorm[2] += (du - duh).dot(du - duh) * weight;
     }
   }


 for(i = 0; i < 3; i++)
   {
   hnorm[i] = sqrt(norm[i] + dnorm[i]);

   norm[i]  = sqrt( norm[i]);
   dnorm[i] = sqrt(dnorm[i]);
   }

 cout.setf(ios::scientific);
 cout.width(12);

 cout << 
     "\n                  L-2 norm      H-1 semi-norm     H-1  norm" << endl;
 cout << "Exact Solution   " << norm[0] << "   " << dnorm[0] 
      << "     " << hnorm[0]  << endl;
 cout << "F. E. Solution   " << norm[1] << "   " << dnorm[1] 
      << "     " << hnorm[1] << endl;
 cout << "Errors           " << norm[2] << "   " << dnorm[2] 
      << "     " << hnorm[2]  << endl;

 cout.unsetf(ios::scientific);

 return;

 error:

 cerr << "Error in norm() -- zero Jacobian" << endl;
 throw("Zero Jacobian");
}

void normTrunc(feMV::vector& uu, feMesh &mm)
{
  int i, ig, n;
 dvector dphidx[Nphi], xx, du, duh;
 dmatrix dsdx, dxds;

 double det, u, uh, ue[Nphi], weight;
 double norm[3], dnorm[3], hnorm[3];

 int nint = 0;
 double factor = 0.01, amin = 1.0e50, amax = 0.0, atot = 0.0, aint = 0.0;
 vector<double> area;

 for(i = 0; i < 3; i++)
   {
   norm[i] = 0.0;
   dnorm[i] = 0.0;
   }
 
 n = 0;

 for (gmMesh::geiterator geit = mm.gebegin(); 
      geit != mm.geend(); geit++ )
 {
     gelement ee = *geit;     
     
     area.push_back(0.0);
     
     for(ig = 0; ig < Ng ; ig++)
     {
	 // Calculate the Transormation Matrix
	 
	 dxds.zero();
	 
	 for(i = 0; i < Nchi; i++)
	 {
	     dxds += dmatrix(ee.xx[i], Dchi[ig][i]);  // outer product
	 }
	 
	 dsdx = dxds.inverse(det);
	 
	 if(det == 0.0) goto error;
	 
	 weight = fabs(det) * Gw[ig];
	 
	 area[n] += weight;	
     }
     
     atot += area[n];
     
     if(area[n] < amin) amin = area[n];
     if(area[n] > amax) amax = area[n];
     n++;
 }
 
 n = 0;

 for (feMesh::feiterator feit = mm.febegin(); 
      feit != mm.feend(); feit++ )
   {
   felement ee = *feit;
     
   if(area[n] > factor * amax)
     {
	 //gelement ee = mm.element(n);
	 
	 nint ++; 
	 aint += area[n];
	 
	 for(ig = 0; ig < Ng ; ig++)
	 {
	     // Calculate the Transormation Matrix
	     
	     xx.zero();
	     dxds.zero();
	     
	     for(i = 0; i < Nchi; i++)
	     {
		 xx += Chi[ig][i] * ee.xx[i];
		 
		 dxds += dmatrix(ee.xx[i], Dchi[ig][i]);  // outer product
	     }
	     
	     dsdx = dxds.inverse(det);
	     
	     if(det == 0.0) goto error;
	     
	     weight = fabs(det) * Gw[ig];
	     
	     dsdx = dsdx.transpose();
	     
	     for(i = 0; i < Nphi; i++)  dphidx[i] = dsdx * Dphi[ig][i];
	     
	     u = exact(xx, du);                  /* Get the exact solution.  */
	     
	     uh = 0.0;
	     duh.zero();
	     
	     uu.extract(ee.ndof(), ee.dof, ue);
	     
	     for(i = 0; i < Nphi; i++)
	     {
		 uh  += ue[i] * Phi[ig][i];
		 duh += ue[i] * dphidx[i];
	     }

	     norm[0] += u * u * weight;
	     norm[1] += uh*uh * weight;
	     norm[2] += (u-uh)*(u-uh) * weight;
	     
	     dnorm[0] +=  du.dot(du)  * weight;
	     dnorm[1] += duh.dot(duh) * weight;
	     dnorm[2] += (du - duh).dot(du - duh) * weight;
	 }
     }
     n++;
 }
 
 for(i = 0; i < 3; i++)
 {
	 hnorm[i] = sqrt(norm[i] + dnorm[i]);
	 
	 norm[i]  = sqrt( norm[i]);
	 dnorm[i] = sqrt(dnorm[i]);
 }
 
 cout.setf(ios::scientific);
 cout.width(12);

 cout << endl;
 cout << "Norms on elements with |K| > " << factor << " (max |K|)" << endl;
 cout << "Volume  = " << atot << ", Volume of fat elements = " << aint << endl;
 cout << "Max |K| = " << amax << ", Min |K| = " << amin << endl;
 cout << "Number of elements discarded = " << n - nint << endl;
 cout << 
     "\n                  L-2 norm      H-1 semi-norm     H-1  norm" << endl;
 cout << "Exact Solution   " << norm[0] << "   " << dnorm[0] 
      << "     " << hnorm[0]  << endl;
 cout << "F. E. Solution   " << norm[1] << "   " << dnorm[1] 
      << "     " << hnorm[1] << endl;
 cout << "Errors           " << norm[2] << "   " << dnorm[2] 
      << "     " << hnorm[2]  << endl;

 cout.unsetf(ios::scientific);

 return;

 error:

 cerr << "Error in norm() -- zero Jacobian" << endl;
 throw("Zero Jacobian");
}

void initializeBasis()
{
  int ig;

#if(NDIM == 2)
  double *Gp1, *Gw1;

  if(Ngb == 3)     {Gp1 = GP3; Gw1 = GW3;}
  else if(Ngb == 5){Gp1 = GP5; Gw1 = GW5;}
  else if(Ngb == 7){Gp1 = GP7; Gw1 = GW7;}
  else
    assert(false);           // others not yet implemented

  for(int i = 0; i < Ngb; i++)
    {
    Gpb[i] = bvector(Gp1[i]);
    Gwb[i] = Gw1[i];

    line2(Gp1[i], Chib[i],  Dchib[i]);       // affine elements

    if(Nphib == 2)      line2(Gp1[i], Phib[i],  Dphib[i]);
    else if(Nphib == 3) line3(Gp1[i], Phib[i],  Dphib[i]);
    else assert(false);
    }

  assert(Ng == 7);

  for(ig = 0; ig < Ng; ig++)
    {
    Gp[ig] = GpTri7[ig];
    Gw[ig] = GwTri7[ig];
      
    tri3(Gp[ig],Chi[ig],Dchi[ig]);           // affine elements
    
    if(Nphi == 3)      tri3(Gp[ig],Phi[ig],Dphi[ig]);
    else if(Nphi == 6) tri6(Gp[ig],Phi[ig],Dphi[ig]);
    else assert(false);
    } 

#elif(NDIM == 3)

  for(ig = 0; ig < Ngb; ig++)
    {
      if(Ngb == 7)
	{
	  Gpb[ig] = GpTri7[ig];
	  Gwb[ig] = GwTri7[ig];
	}
      else
	assert(false);

      tri3(Gpb[ig],Chib[ig],Dchib[ig]);      // affine elements

      if(Nphib == 3)      tri3(Gpb[ig],Phib[ig],Dphib[ig]);
      else if(Nphib == 6) tri6(Gpb[ig],Phib[ig],Dphib[ig]);
      else assert(false);
    }

  assert(Ng == 14);

  for(ig = 0; ig < Ng; ig++)
    {
      if(Ng == 14)
	{
	  Gp[ig] = GpTet14[ig];
	  Gw[ig] = GwTet14[ig];
	}
      else
	assert(false);

      tet4( Gp[ig],Chi[ig],Dchi[ig]);       // affine elements
      
      if(Nphi == 4)       tet4( Gp[ig],Phi[ig],Dphi[ig]);
      else if(Nphi == 10) tet10(Gp[ig],Phi[ig],Dphi[ig]);
      else assert(false);
    }

#else
  assert(false);
#endif

return;
}

void output(feMV::vector& u)
{
  int j, i=0, n = u.length();

 if(n > 25) return;

 cout << "\nSolution u(i), i = 0,1, ... ," << n-1;

 cout.setf(ios::fixed);

 while( i < n )
   {
   cout << "\ni = " << setw(3) << i;

   for(j = 0; (j < 5) && (i <= n); j++ ) cout << "  " << setw(10) << u[i++];
   }

 cout << endl;

 cout.unsetf(ios::fixed);

 return;
}

