// Sample program to solve Poisson's equation
// Sample program to solve Poisson's equation
// Sample program to solve Poisson's equation

#include<sstream>
#include<fstream>
#include<iostream>
#include<iomanip>
#include<cstdio>
#include<cmath>
#include<cassert>
#include<ctime>

using namespace std;

#include"FiniteElement/svector.C"
#include"FiniteElement/smatrix.C"
#include"FiniteElement/sbmatrix.C"

// options specified next -- (un)comment options as required
// options specified next -- (un)comment options as required
// options specified next -- (un)comment options as required

// The DeBug flag enables some (possibly slow) checks such
// as array bound checking.

//#define DeBug

// Two or Three dimensions (2d code reads triangle mesh)

// #define NDIM 2
#define NDIM 3

// interface to mesher dir3

#define TPLCMESH
//#define PLCMESH

// finite element basis functions

//#define LinearElements
#define QuadraticElements

// Linear solver

#define BandMatrix
// #define SpoolesMatrix
// #define PetscMatrix

// end option specification -- code starts
// end option specification -- code starts
// end option specification -- code starts

const int Ndim = NDIM;             // two or three dimensional code

const int Ng1  = 7;                // 1d quadrature rule

// Define dvector/dmatrix to be the small (Ndim) vector/matrix class

typedef svector<Ndim> dvector;
typedef smatrix<Ndim> dmatrix;
typedef sbmatrix<Ndim> bmatrix;
typedef svector<Ndim-1> bvector;    // boundary Gauss points

// Set up mesh; 2d files from triangle output, 3d from dir3

#if NDIM == 2                
  const int Ngb = Ng1;             // One dimensional boundary integrals
  const int Ng  = 7;               // Two dimensional area     integrals

  const int Nchib = 2;             // 2 = linear geometry;
  const int Nchi  = 3;             // 3 = linear geometry;

  #ifdef LinearElements
    const int Nphib  = 2;          // Linear elements
    const int Nphi   = 3;
  #elif defined QuadraticElements
    const int Nphib  = 3;          // Quadratic elements
    const int Nphi   = 6;
  #endif

  #include"MeshGeneration/triangleMesh.C"
  typedef triangleMesh gmMesh;
#endif

#if NDIM == 3
  const int Ngb   = 7;              // Two   dimensional boundary integrals
  const int Ng    = 14;             // Three dimensional volume   integrals

  const int Nchib = 3;              // 3 = linear geometry
  const int Nchi  = 4;              // 4 = linear geometry

  #ifdef LinearElements
    const int Nphib = 3;            // Linear elements
    const int Nphi  = 4;            
  #elif defined QuadraticElements
    const int Nphib = 6;            // Quadratic elements
    const int Nphi  = 10;            
  #endif

  #ifdef TPLCMESH                   // simple interface to read .tplc file  
    #include"MeshGeneration/dir3MeshTPLC.C"
  #elif defined PLCMESH             // dir3 constructs mesh from .plc file
    #include"MeshGeneration/dir3MeshPLC.C"
  #endif

  typedef dir3Mesh gmMesh;
#endif

#include"FiniteElement/basisFns.C"      // basis functions
#include"FiniteElement/feMesh.C"      // feMesh class 

// Arrays to store basis functions evaluated at the Gauss points

dvector Gp[Ng]; double Gw[Ng];          // Gauss points and weights in space
double Chi[Ng][Nchi]; dvector Dchi[Ng][Nchi];       // geometry basis fns
double Phi[Ng][Nphi]; dvector Dphi[Ng][Nphi];       // soln basis fns

bvector Gpb[Ngb]; double Gwb[Ngb];     // Boundary Gauss points and weights 
double Chib[Ngb][Nchib]; bvector Dchib[Ngb][Nchib];   // geometry basis fns
double Phib[Ngb][Nphib]; bvector Dphib[Ngb][Nphib];   // basis fns for soln

// Get the big matrix/vector class

class FeMdata    // parameters a matrix/linear solver constructor may need
{
 public:
  int nvar;      // size of the matrix/vectors
  int nzero;     // estimated number of zeros per row
  int band;      // bandwidth (for banded solvers)
  bool pivot;    // pivoting required for direct solvers
  bool pcflag;   // seperate preconditioner for itterative solvers
  double tol;    // tolerence for itterative solvers
  FeMdata() 
    : nvar(0), nzero(0), band(0), pivot(false), pcflag(false), tol(1.0e-10)
    {return;}
};

#ifdef BandMatrix
  #include"LinearAlgebra/bandMatrix.C"
  typedef bandMV feMV;
#endif

#ifdef SpoolesMatrix
  #include"LinearAlgebra/spoolesMatrix.C"
  typedef spoolesMV feMV;
#endif

#ifdef PetscMatrix

// petsc 2.3
//#define PETSC23 

// petsc 3.0
#define PETSC30

  #include"LinearAlgebra/petscMatrix.C"
  typedef petscMV feMV;
#endif

// Set up Gauss quadrature points

double GP3[] = {-0.774596669241483,0.0,0.774596669241483};
double GW3[] = {0.555555555555556,0.88888888888889,0.555555555555556};

double GP5[] = {-0.906179845938664,-0.538469310105683,0.0,
                0.906179845938664, 0.538469310105683};
double GW5[] = {0.236926885056189, 0.478628670499366,0.568888888888889,
                0.236926885056189, 0.478628670499366};

double GP7[] = {-0.949107912342759, -0.741531185599394, -0.405845151377397,
		 0.000000000000000,
		 0.405845151377397,  0.741531185599394, 0.949107912342759};

double GW7[] = {0.129484966168870, 0.279705391489277, 0.381830050505119,
		0.417959183673469,
		0.381830050505119, 0.279705391489277, 0.129484966168870};

// Seven point quadrature rule for the trinagle
// Note the Gauss points are  B =  2/7 (+ or -) sqrt(15)/21
// and  the Gauss weights are W = 31/240 (+ -)  sqrt(15)/1200
// The center point and weight below are exact

double BB1=0.4701420641051151, AA1=1.0-2.0*BB1, W1=0.1323941527885062/2.0;
double BB2=0.1012865073234563, AA2=1.0-2.0*BB2, W2=0.1259391805448272/2.0;

svector<2> GpTri7[] = {svector<2>(1.0/3.0, 1.0/3.0),
    svector<2>(AA1,BB1), svector<2>(BB1,AA1), svector<2>(BB1,BB1), 
    svector<2>(AA2,BB2), svector<2>(BB2,AA2), svector<2>(BB2,BB2)};

double GwTri7[] = {0.1125, W1,W1,W1, W2,W2,W2};

// 14 point quadrature rule for the tetrahedra

double a1 = 0.31088591926330060980,  b1 = 1.0 - 3.0*a1; 
double a2 = 0.092735250310891226402, b2 = 1.0 - 3.0*a2;
double a3 = 0.045503704125649649492, b3 = 0.5 - a3;
double w1 = 0.018781320953002641800;
double w2 = 0.012248840519393658257;
double w3 = 0.0070910034628469110730;

svector<3> GpTet14[] = 
  {svector<3>(a1,a1,a1), svector<3>(a1,a1,b1), 
   svector<3>(a1,b1,a1), svector<3>(b1,a1,a1),
   svector<3>(a2,a2,a2), svector<3>(a2,a2,b2), 
   svector<3>(a2,b2,a2), svector<3>(b2,a2,a2),
   svector<3>(a3,a3,b3), svector<3>(a3,b3,a3), svector<3>(b3,a3,a3),
   svector<3>(b3,b3,a3), svector<3>(b3,a3,b3), svector<3>(a3,b3,b3)};

double GwTet14[] = {w1,w1,w1,w1, w2,w2,w2,w2, w3,w3,w3, w3,w3,w3};

// Function Prototypes

void elment(dvector xy[Nphi], double a[Nphi][Nphi] , double f[Nphi]);
void bndyc(bfelement be, double ab[Nphib][Nphib], double fb[Nphib]);
double dirichlet(dvector xx);
double neumann(dvector xx, dvector normal);


void norm(feMV::vector& uu, feMesh &mm);
void normTrunc(feMV::vector& uu, feMesh &mm);

void initializeBasis();
void output(feMV::vector& u);

void line2(double eta, double psi[2], svector<1> dpsi[2]);
void line3(double eta, double psi[3], svector<1> dpsi[3]);

void line1(double eta, double psi[1], double dpsi[1]);
void line2(double eta, double psi[2], double dpsi[2]);

void tri3(svector<2> xi, double psi[3], svector<2> dpsi[3]);
void tri6(svector<2> xi, double psi[6], svector<2> dpsi[6]);

void tet4( const svector<3> xi, double phi[4] , svector<3> dphi[4]);
void tet10(const svector<3> xi, double phi[10], svector<3> dphi[10]);

// STRUCTURE TO CONTAIN DATA ROUTINES FOR DIFFERENT EXAMPLES

typedef struct
  {
    /* Names:
       exact   = exact solution
       rhs     = right hand side
       penalty = penalty function
       desc    = description
    */

    void (*initialize)(void);
    double (*exact)(dvector xx, dvector &du);
    void   (*coeff)(dvector xx, dmatrix &A, double &b);
    double (*rhs)(dvector xx);
    string desc;
  } example_data;
                                   
#if NDIM == 2

  const int EgMax = 1;

  #include"examples2d.C"     

  example_data Examples[] =   
    {
      {initialize0, exact0, coeff0, rhs0, "Laplaces Equation"},
      {initialize1, exact1, coeff1, rhs1, "Poissons Equation"}
    };
#endif

#if NDIM == 3

  const int EgMax = 1;

  #include"examples3d.C"     

  example_data Examples[] =   
    {
      {initialize0, exact0, coeff0, rhs0, "Laplaces Equation"},
      {initialize1, exact1, coeff1, rhs1, "Poissons Equation"}
    };
#endif

